//! Stream buffering for sync of encrypted socked.
use super::*;
use smallvec::SmallVec;
use std::io;
use std::{
    task::{
	Poll, Context,
    },
    pin::Pin,
};
use bytes::{
    Buf, BufMut,
};

/// A wrapping buffer over a writer and/or reader.
#[pin_project]
#[derive(Debug)]
pub struct Buffered<T: ?Sized, const SIZE: usize>
{
    /// Current internal buffer
    /// When it's full to `SIZE`, it should be written to `stream` at once then cleared when it's been written.
    buffer: SmallVec<[u8; SIZE]>, //TODO: Can we have a non-spilling stack vec?
    #[pin] stream: T
}

impl<T, const SIZE: usize> Buffered<T, SIZE>
{
    /// Create a new staticly sized buffer wrapper around this stream
    pub fn new(stream: T) -> Self
    {
	Self {
	    buffer: SmallVec::new(),
	    stream,
	}
    }
    /// Consume into the wrapped stream
    pub fn into_inner(self) -> T
    {
	self.stream
    }
    /// The inner stream
    pub fn inner(&self) -> &T
    {
	&self.stream
    }
    /// A mutable reference to the backing stream
    pub fn inner_mut(&mut self) -> &mut T
    {
	&mut self.stream
    }

    /// The current buffer bytes.
    pub fn current_buffer(&self) -> &[u8]
    {
	&self.buffer[..]
    }

    /// Is the current internal buffer empty?
    ///
    /// You can flush a partially-filled buffer to the backing stream of a writer with `.flush().await`.
    pub fn is_empty(&self) -> bool
    {
	self.buffer.is_empty()
    }
}

#[inline] fn div_mod<V>(a: V, b: V) -> (V, <V as std::ops::Div>::Output, <V as std::ops::Rem>::Output)
where V: std::ops::Div + std::ops::Rem + Clone
{
    (a.clone(), a.clone() / b.clone(), a % b)
}

// XXX: I don't think writing futures like this is safe. Expand the inline `async{}`s into actual polling.
impl<T: ?Sized + Unpin, const SIZE: usize> AsyncWrite for Buffered<T, SIZE>
where T: AsyncWrite
{
    fn poll_write(self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &[u8]) -> Poll<Result<usize, io::Error>> {
	// TODO: Don't write poll methods like this ffs... Write it properly.
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    let mut written=0;
	    let mut err = None;
	    this.buffer.extend_from_slice(buf);

	    for chunk in this.buffer.chunks_exact(SIZE)
	    {
		match this.stream.write_all(&chunk).await {
		    Ok(()) => {
			written += chunk.len();
		    },
		    Err(e) => {
			err = Some(e);
			break;
		    },
		}
	    }
	    this.buffer.drain(0..written);
	    if let Some(err) = err {
		Err(err)
	    } else {
		Ok(written)
	    }
	};
	tokio::pin!(fut);
	fut.poll(cx)
	    
    }
    fn poll_flush(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    let wres = if this.buffer.len() > 0 {
		this.stream.write_all(&this.buffer[..]).await
	    } else {
		Ok(())
	    };
	    this.stream.flush().await?;
	    wres
	};
	tokio::pin!(fut);
	fut.poll(cx)
    }
    fn poll_shutdown(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), io::Error>> {
	let this = self.get_mut();
	let fut = async {
	    use tokio::prelude::*;
	    
	    this.flush().await?;
	    this.stream.shutdown().await
	};
	tokio::pin!(fut);
	fut.poll(cx)
    }
}


// XXX: I don't think writing futures like this is safe. Expand the inline `async{}`s into actual polling.
impl<T: ?Sized + Unpin, const SIZE: usize> AsyncRead for Buffered<T, SIZE>
where T: AsyncRead
{
    fn poll_read(self: Pin<&mut Self>, cx: &mut Context<'_>, buf: &mut [u8]) -> Poll<io::Result<usize>> {
	todo!("How to properly do this?")
    }
}

//TODO: Write tests for writer
